// Copyright 2003 Adam Megacz, see the COPYING file for licensing [GPL]
package org.xwt.js;

import org.xwt.util.*;
import java.io.*;

/**
 *  Parses a stream of lexed tokens into a tree of CompiledFunctionImpl's.
 *
 *  There are three kinds of things we parse: blocks, statements, and
 *  expressions.
 *
 *  - Expressions are a special type of statement that evaluates to a
 *    value (for example, "break" is not an expression, * but "3+2"
 *    is).  Some tokens sequences start expressions (for * example,
 *    literal numbers) and others continue an expression which * has
 *    already been begun (for example, '+').  Finally, some *
 *    expressions are valid targets for an assignment operation; after
 *    * each of these expressions, continueExprAfterAssignable() is
 *    called * to check for an assignment operation.
 *
 *  - A statement ends with a semicolon and does not return a value.
 *
 *  - A block is a single statement or a sequence of statements
 *    surrounded by curly braces.
 *
 *  Each parsing method saves the parserLine before doing its actual
 *  work and restores it afterwards.  This ensures that parsing a
 *  subexpression does not modify the line number until a token
 *  *after* the subexpression has been consumed by the parent
 *  expression.
 *
 *  Technically it would be a better design for this class to build an
 *  intermediate parse tree and use that to emit bytecode.  Here's the
 *  tradeoff:
 *
 *  Advantages of building a parse tree:
 *  - easier to apply optimizations
 *  - would let us handle more sophisticated languages than JavaScript
 *
 *  Advantages of leaving out the parse tree
 *  - faster compilation
 *  - less load on the garbage collector
 *  - much simpler code, easier to understand
 *  - less error-prone
 *
 *  Fortunately JS is such a simple language that we can get away with
 *  the half-assed approach and still produce a working, complete
 *  compiler.
 *
 *  The bytecode language emitted doesn't really cause any appreciable
 *  semantic loss, and is itself a parseable language very similar to
 *  Forth or a postfix variant of LISP.  This means that the bytecode
 *  can be transformed into a parse tree, which can be manipulated.
 *  So if we ever want to add an optimizer, it could easily be done by
 *  producing a parse tree from the bytecode, optimizing that tree,
 *  and then re-emitting the bytecode.  The parse tree node class
 *  would also be much simpler since the bytecode language has so few
 *  operators.
 *
 *  Actually, the above paragraph is slightly inaccurate -- there are
 *  places where we push a value and then perform an arbitrary number
 *  of operations using it before popping it; this doesn't parse well.
 *  But these cases are clearly marked and easy to change if we do
 *  need to move to a parse tree format.
 */
class Parser extends Lexer implements ByteCodes {


    // Constructors //////////////////////////////////////////////////////

    public Parser(Reader r, String sourceName, int line) throws IOException { super(r, sourceName, line); }

    /** for debugging */
    public static void main(String[] s) throws Exception {
        CompiledFunctionImpl block = new JS.CompiledFunction("stdin", 0, new InputStreamReader(System.in), null);
        if (block == null) return;
        System.out.println(block);
    }


    // Statics ////////////////////////////////////////////////////////////

    static byte[] precedence = new byte[MAX_TOKEN + 1];
    static boolean[] isRightAssociative = new boolean[MAX_TOKEN + 1];
    // Use this as the precedence when we want anything up to the comma
    private final static int NO_COMMA = 2;
    static {
        isRightAssociative[ASSIGN] =
            isRightAssociative[ASSIGN_BITOR] =
            isRightAssociative[ASSIGN_BITXOR] =
            isRightAssociative[ASSIGN_BITAND] =
            isRightAssociative[ASSIGN_LSH] =
            isRightAssociative[ASSIGN_RSH] =
            isRightAssociative[ASSIGN_URSH] =
            isRightAssociative[ASSIGN_ADD] =
            isRightAssociative[ASSIGN_SUB] =
            isRightAssociative[ASSIGN_MUL] =
            isRightAssociative[ASSIGN_DIV] =
            isRightAssociative[ASSIGN_MOD] = true;

        precedence[COMMA] = 1;
        // 2 is intentionally left unassigned. we use minPrecedence==2 for comma separated lists
        precedence[ASSIGN] =
            precedence[ASSIGN_BITOR] =
            precedence[ASSIGN_BITXOR] =
            precedence[ASSIGN_BITAND] =
            precedence[ASSIGN_LSH] =
            precedence[ASSIGN_RSH] =
            precedence[ASSIGN_URSH] =
            precedence[ASSIGN_ADD] =
            precedence[ASSIGN_SUB] =
            precedence[ASSIGN_MUL] =
            precedence[ASSIGN_DIV] =
            precedence[ASSIGN_MOD] = 3;
        precedence[HOOK] = 4;
        precedence[OR] = 5;
        precedence[AND] = 6;
        precedence[BITOR] = 7;
        precedence[BITXOR] = 8;
        precedence[BITAND] = 9;
        precedence[EQ] = precedence[NE] = precedence[SHEQ] = precedence[SHNE] = 10;
        precedence[LT] = precedence[LE] = precedence[GT] = precedence[GE] = 11;
        precedence[LSH] = precedence[RSH] = precedence[URSH] = 12;
        precedence[ADD] = precedence[SUB] = 12;
        precedence[MUL] = precedence[DIV] = precedence[MOD] = 13;
        precedence[BITNOT] =  precedence[BANG] = precedence[TYPEOF] = 14;
        precedence[DOT] = precedence[LB] = precedence[LP] =  precedence[INC] = precedence[DEC] = 15;
    }


    // Parsing Logic /////////////////////////////////////////////////////////

    /** gets a token and throws an exception if it is not <tt>code</tt> */
    private void consume(int code) throws IOException {
        if (getToken() != code) throw pe("expected " + codeToString[code] + ", got " + (op == -1 ? "EOF" : codeToString[op]));
    }

    /**
     *  Parse the largest possible expression containing no operators
     *  of precedence below <tt>minPrecedence</tt> and append the
     *  bytecodes for that expression to <tt>appendTo</tt>; the
     *  appended bytecodes MUST grow the stack by exactly one element.
     */ 
    private void startExpr(CompiledFunctionImpl appendTo, int minPrecedence) throws IOException {
        int saveParserLine = parserLine;
        _startExpr(appendTo, minPrecedence);
        parserLine = saveParserLine;
    }
    private void _startExpr(CompiledFunctionImpl appendTo, int minPrecedence) throws IOException {
        int tok = getToken();
        CompiledFunctionImpl b = appendTo;

        switch (tok) {
        case -1: throw pe("expected expression");

        // all of these simply push values onto the stack
        case NUMBER: b.add(parserLine, LITERAL, number); break;
        case STRING: b.add(parserLine, LITERAL, string); break;
        case THIS: b.add(parserLine, TOPSCOPE, null); break;
        case NULL: b.add(parserLine, LITERAL, null); break;
        case TRUE: case FALSE: b.add(parserLine, LITERAL, new Boolean(tok == TRUE)); break;

        case LB: {
            b.add(parserLine, ARRAY, new Integer(0));                       // push an array onto the stack
            int size0 = b.size();
            int i = 0;
            if (peekToken() != RB)
                while(true) {                                               // iterate over the initialization values
                    int size = b.size();
                    b.add(parserLine, LITERAL, new Integer(i++));           // push the index in the array to place it into
                    if (peekToken() == COMMA || peekToken() == RB)
                        b.add(parserLine, LITERAL, null);                   // for stuff like [1,,2,]
                    else
                        startExpr(b, NO_COMMA);                             // push the value onto the stack
                    b.add(parserLine, PUT);                                 // put it into the array
                    b.add(parserLine, POP);                                 // discard the value remaining on the stack
                    if (peekToken() == RB) break;
                    consume(COMMA);
                }
            b.set(size0 - 1, new Integer(i));                               // back at the ARRAY instruction, write the size of the array
            consume(RB);
            break;
        }
        case SUB: {  // negative literal (like "3 * -1")
            consume(NUMBER);
            b.add(parserLine, LITERAL, new Double(number.doubleValue() * -1));
            break;
        }
        case LP: {  // grouping (not calling)
            startExpr(b, -1);
            consume(RP);
            break;
        }
        case INC: case DEC: {  // prefix (not postfix)
            startExpr(b, precedence[tok]);
            int prev = b.size() - 1;
            if (b.get(prev) == GET && b.getArg(prev) != null)
                b.set(prev, LITERAL, b.getArg(prev));
            else if(b.get(prev) == GET)
                b.pop();
            else
                throw pe("prefixed increment/decrement can only be performed on a valid assignment target");
            b.add(parserLine, tok, Boolean.TRUE);
            break;
        }
        case BANG: case BITNOT: case TYPEOF: {
            startExpr(b, precedence[tok]);
            b.add(parserLine, tok);
            break;
        }
        case LC: { // object constructor
            b.add(parserLine, OBJECT, null);                                     // put an object on the stack
            if (peekToken() != RC)
                while(true) {
                    if (peekToken() != NAME && peekToken() != STRING)
                        throw pe("expected NAME or STRING");
                    getToken();
                    b.add(parserLine, LITERAL, string);                          // grab the key
                    consume(COLON);
                    startExpr(b, NO_COMMA);                                      // grab the value
                    b.add(parserLine, PUT);                                      // put the value into the object
                    b.add(parserLine, POP);                                      // discard the remaining value
                    if (peekToken() == RC) break;
                    consume(COMMA);
                    if (peekToken() == RC) break;                                // we permit {,,} -- I'm not sure if ECMA does
                }
            consume(RC);
            break;
        }
        case NAME: {
            b.add(parserLine, TOPSCOPE);
            b.add(parserLine, LITERAL, string);
            continueExprAfterAssignable(b,minPrecedence);
            break;
        }
        case FUNCTION: {
            consume(LP);
            int numArgs = 0;
            CompiledFunctionImpl b2 = new JS.CompiledFunction(sourceName, parserLine, null, null);
            b.add(parserLine, NEWFUNCTION, b2);

            // function prelude; arguments array is already on the stack
            b2.add(parserLine, TOPSCOPE);
            b2.add(parserLine, SWAP);
            b2.add(parserLine, DECLARE, "arguments");                     // declare arguments (equivalent to 'var arguments;')
            b2.add(parserLine, SWAP);                                     // set this.arguments and leave the value on the stack
            b2.add(parserLine, PUT);

            while(peekToken() != RP) {                                    // run through the list of argument names
                if (peekToken() == NAME) {
                    consume(NAME);                                        // a named argument
                    String varName = string;
                    
                    b2.add(parserLine, DUP);                              // dup the args array 
                    b2.add(parserLine, GET, new Integer(numArgs));   // retrieve it from the arguments array
                    b2.add(parserLine, TOPSCOPE);
                    b2.add(parserLine, SWAP);
                    b2.add(parserLine, DECLARE, varName);                  // declare the name
                    b2.add(parserLine, SWAP);
                    b2.add(parserLine, PUT); 
                    b2.add(parserLine, POP);                               // pop the value
                    b2.add(parserLine, POP);                               // pop the scope                  
                }
                if (peekToken() == RP) break;
                consume(COMMA);
                numArgs++;
            }
            consume(RP);

            b2.add(parserLine, POP);                                      // pop off the arguments array
            b2.add(parserLine, POP);                                      // pop off TOPSCOPE
            
           if(peekToken() != LC)
                throw pe("Functions must have a block surrounded by curly brackets");
                
            parseBlock(b2, null);                                   // the function body

            b2.add(parserLine, LITERAL, null);                        // in case we "fall out the bottom", return NULL
            b2.add(parserLine, RETURN);

            break;
        }
        default: throw pe("expected expression, found " + codeToString[tok] + ", which cannot start an expression");
        }

        // attempt to continue the expression
        continueExpr(b, minPrecedence);
    }


    /**
     *  Assuming that a complete assignable (lvalue) has just been
     *  parsed and the object and key are on the stack,
     *  <tt>continueExprAfterAssignable</tt> will attempt to parse an
     *  expression that modifies the assignable.  This method always
     *  decreases the stack depth by exactly one element.
     */
    private void continueExprAfterAssignable(CompiledFunctionImpl b,int minPrecedence) throws IOException {
        int saveParserLine = parserLine;
        _continueExprAfterAssignable(b,minPrecedence);
        parserLine = saveParserLine;
    }
    private void _continueExprAfterAssignable(CompiledFunctionImpl b,int minPrecedence) throws IOException {
        if (b == null) throw new Error("got null b; this should never happen");
        int tok = getToken();
        if (minPrecedence != -1 && (precedence[tok] < minPrecedence || (precedence[tok] == minPrecedence && !isRightAssociative[tok])))
            // force the default case
            tok = -1;
        switch(tok) {
        case ASSIGN_BITOR: case ASSIGN_BITXOR: case ASSIGN_BITAND: case ASSIGN_LSH: case ASSIGN_RSH: case ASSIGN_URSH:
        case ASSIGN_ADD: case ASSIGN_SUB: case ASSIGN_MUL: case ASSIGN_DIV: case ASSIGN_MOD: {
            b.add(parserLine, GET_PRESERVE);
            startExpr(b,  precedence[tok]);
            // tok-1 is always s/^ASSIGN_// (0 is BITOR, 1 is ASSIGN_BITOR, etc) 
            b.add(parserLine, tok - 1, tok-1==ADD ? new Integer(2) : null);
            b.add(parserLine, PUT);
            b.add(parserLine, SWAP);
            b.add(parserLine, POP);
            break;
        }
        case INC: case DEC: { // postfix
            b.add(parserLine, tok, Boolean.FALSE);
            break;
        }
        case ASSIGN: {
            startExpr(b, precedence[tok]);
            b.add(parserLine, PUT);
            b.add(parserLine, SWAP);
            b.add(parserLine, POP);
            break;
        }
        case LP: {
            int n = parseArgs(b);
            b.add(parserLine,CALLMETHOD,new Integer(n));
            break;
        }
        default: {
            pushBackToken();
            if(b.get(b.size()-1) == LITERAL && b.getArg(b.size()-1) != null)
                b.set(b.size()-1,GET,b.getArg(b.size()-1));
            else
                b.add(parserLine, GET);
            return;
        }
        }
    }


    /**
     *  Assuming that a complete expression has just been parsed,
     *  <tt>continueExpr</tt> will attempt to extend this expression by
     *  parsing additional tokens and appending additional bytecodes.
     *
     *  No operators with precedence less than <tt>minPrecedence</tt>
     *  will be parsed.
     *
     *  If any bytecodes are appended, they will not alter the stack
     *  depth.
     */
    private void continueExpr(CompiledFunctionImpl b, int minPrecedence) throws IOException {
        int saveParserLine = parserLine;
        _continueExpr(b, minPrecedence);
        parserLine = saveParserLine;
    }
    private void _continueExpr(CompiledFunctionImpl b, int minPrecedence) throws IOException {
        if (b == null) throw new Error("got null b; this should never happen");
        int tok = getToken();
        if (tok == -1) return;
        if (minPrecedence != -1 && (precedence[tok] < minPrecedence || (precedence[tok] == minPrecedence && !isRightAssociative[tok]))) {
            pushBackToken();
            return;
        }

        switch (tok) {
        case LP: {  // invocation (not grouping)
            int n = parseArgs(b);
            b.add(parserLine, CALL, new Integer(n));
            break;
        }
        case BITOR: case BITXOR: case BITAND: case SHEQ: case SHNE: case LSH:
        case RSH: case URSH: case MUL: case DIV: case MOD:
        case GT: case GE: case EQ: case NE: case LT: case LE: case SUB: {
            startExpr(b, precedence[tok]);
            b.add(parserLine, tok);
            break;
        }
        case ADD: {
            int count=1;
            int nextTok;
            do {
                startExpr(b,precedence[tok]);
                count++;
                nextTok = getToken();
            } while(nextTok == tok);
            pushBackToken();
            b.add(parserLine, tok, new Integer(count));
            break;
        }
        case OR: case AND: {
            b.add(parserLine, tok == AND ? b.JF : b.JT, new Integer(0));       // test to see if we can short-circuit
            int size = b.size();
            startExpr(b, precedence[tok]);                                     // otherwise check the second value
            b.add(parserLine, JMP, new Integer(2));                            // leave the second value on the stack and jump to the end
            b.add(parserLine, LITERAL, tok == AND ?
                  new Boolean(false) : new Boolean(true));                     // target of the short-circuit jump is here
            b.set(size - 1, new Integer(b.size() - size));                     // write the target of the short-circuit jump
            break;
        }
        case DOT: {
            consume(NAME);
            b.add(parserLine, LITERAL, string);
            continueExprAfterAssignable(b,minPrecedence);
            break;
        }
        case LB: { // subscripting (not array constructor)
            startExpr(b, -1);
            consume(RB);
            continueExprAfterAssignable(b,minPrecedence);
            break;
        }
        case HOOK: {
            b.add(parserLine, JF, new Integer(0));                // jump to the if-false expression
            int size = b.size();
            startExpr(b, minPrecedence);                          // write the if-true expression
            b.add(parserLine, JMP, new Integer(0));               // if true, jump *over* the if-false expression     
            b.set(size - 1, new Integer(b.size() - size + 1));    // now we know where the target of the jump is
            consume(COLON);
            size = b.size();
            startExpr(b, minPrecedence);                          // write the if-false expression
            b.set(size - 1, new Integer(b.size() - size + 1));    // this is the end; jump to here
            break;
        }
        case COMMA: {
            // pop the result of the previous expression, it is ignored
            b.add(parserLine,POP);
            startExpr(b,-1);
            break;
        }
        default: {
            pushBackToken();
            return;
        }
        }

        continueExpr(b, minPrecedence);                           // try to continue the expression
    }
    
    // parse a set of comma separated function arguments, assume LP has already been consumed
    private int parseArgs(CompiledFunctionImpl b) throws IOException {
        int i = 0;
        while(peekToken() != RP) {
            i++;
            if (peekToken() != COMMA) {
                startExpr(b, NO_COMMA);
                if (peekToken() == RP) break;
            }
            consume(COMMA);
        }
        consume(RP);
        return i;
    }
    
    /** Parse a block of statements which must be surrounded by LC..RC. */
    void parseBlock(CompiledFunctionImpl b) throws IOException { parseBlock(b, null); }
    void parseBlock(CompiledFunctionImpl b, String label) throws IOException {
        int saveParserLine = parserLine;
        _parseBlock(b, label);
        parserLine = saveParserLine;
    }
    void _parseBlock(CompiledFunctionImpl b, String label) throws IOException {
        if (peekToken() == -1) return;
        else if (peekToken() != LC) parseStatement(b, null);
        else {
            consume(LC);
            while(peekToken() != RC && peekToken() != -1) parseStatement(b, null);
            consume(RC);
        }
    }

    /** Parse a single statement, consuming the RC or SEMI which terminates it. */
    void parseStatement(CompiledFunctionImpl b, String label) throws IOException {
        int saveParserLine = parserLine;
        _parseStatement(b, label);
        parserLine = saveParserLine;
    }
    void _parseStatement(CompiledFunctionImpl b, String label) throws IOException {
        int tok = peekToken();
        if (tok == -1) return;
        switch(tok = getToken()) {
            
        case THROW: case ASSERT: case RETURN: {
            if (tok == RETURN && peekToken() == SEMI)
                b.add(parserLine, LITERAL, null);
            else
                startExpr(b, -1);
            b.add(parserLine, tok);
            consume(SEMI);
            break;
        }
        case BREAK: case CONTINUE: {
            if (peekToken() == NAME) consume(NAME);
            b.add(parserLine, tok, string);
            consume(SEMI);
            break;
        }
        case VAR: {
            b.add(parserLine, TOPSCOPE);                         // push the current scope
            while(true) {
                consume(NAME);
                b.add(parserLine, DECLARE, string);               // declare it
                if (peekToken() == ASSIGN) {                     // if there is an '=' after the variable name
                    consume(ASSIGN);
                    startExpr(b, NO_COMMA);
                    b.add(parserLine, PUT);                      // assign it
                    b.add(parserLine, POP);                      // clean the stack
                } else {
                    b.add(parserLine, POP);                      // pop the string pushed by declare
                }   
                if (peekToken() != COMMA) break;
                consume(COMMA);
            }
            b.add(parserLine, POP);                              // pop off the topscope
            if ((mostRecentlyReadToken != RC || peekToken() == SEMI) && peekToken() != -1 && mostRecentlyReadToken != SEMI) consume(SEMI);
            break;
        }
        case IF: {
            consume(LP);
            startExpr(b, -1);
            consume(RP);
            
            b.add(parserLine, JF, new Integer(0));                    // if false, jump to the else-block
            int size = b.size();
            parseStatement(b, null);
            
            if (peekToken() == ELSE) {
                consume(ELSE);
                b.add(parserLine, JMP, new Integer(0));               // if we took the true-block, jump over the else-block
                b.set(size - 1, new Integer(b.size() - size + 1));
                size = b.size();
                parseStatement(b, null);
            }
            b.set(size - 1, new Integer(b.size() - size + 1));        // regardless of which branch we took, b[size] needs to point here
            break;
        }
        case WHILE: {
            consume(LP);
            if (label != null) b.add(parserLine, LABEL, label);
            b.add(parserLine, LOOP);
            int size = b.size();
            b.add(parserLine, POP);                                   // discard the first-iteration indicator
            startExpr(b, -1);
            b.add(parserLine, JT, new Integer(2));                    // if the while() clause is true, jump over the BREAK
            b.add(parserLine, BREAK);
            consume(RP);
            parseStatement(b, null);
            b.add(parserLine, CONTINUE);                              // if we fall out of the end, definately continue
            b.set(size - 1, new Integer(b.size() - size + 1));        // end of the loop
            break;
        }
        case SWITCH: {
            consume(LP);
            if (label != null) b.add(parserLine, LABEL, label);
            b.add(parserLine, LOOP);
            int size0 = b.size();
            startExpr(b, -1);
            consume(RP);
            consume(LC);
            while(true)
                if (peekToken() == CASE) {                         // we compile CASE statements like a bunch of if..else's
                    consume(CASE);
                    b.add(parserLine, DUP);                        // duplicate the switch() value; we'll consume one copy
                    startExpr(b, -1);
                    consume(COLON);
                    b.add(parserLine, EQ);                         // check if we should do this case-block
                    b.add(parserLine, JF, new Integer(0));         // if not, jump to the next one
                    int size = b.size();
                    while(peekToken() != CASE && peekToken() != DEFAULT && peekToken() != RC) parseStatement(b, null);
                    b.set(size - 1, new Integer(1 + b.size() - size));
                } else if (peekToken() == DEFAULT) {
                    consume(DEFAULT);
                    consume(COLON);
                    while(peekToken() != CASE && peekToken() != DEFAULT && peekToken() != RC) parseStatement(b, null);
                } else if (peekToken() == RC) {
                    consume(RC);
                    b.add(parserLine, BREAK);                      // break out of the loop if we 'fall through'
                    break;
                } else {
                    throw pe("expected CASE, DEFAULT, or RC; got " + codeToString[peekToken()]);
                }
            b.set(size0 - 1, new Integer(b.size() - size0 + 1));      // end of the loop
            break;
        }
            
        case DO: {
            if (label != null) b.add(parserLine, LABEL, label);
            b.add(parserLine, LOOP);
            int size = b.size();
            parseStatement(b, null);
            consume(WHILE);
            consume(LP);
            startExpr(b, -1);
            b.add(parserLine, JT, new Integer(2));                  // check the while() clause; jump over the BREAK if true
            b.add(parserLine, BREAK);
            b.add(parserLine, CONTINUE);
            consume(RP);
            consume(SEMI);
            b.set(size - 1, new Integer(b.size() - size + 1));      // end of the loop; write this location to the LOOP instruction
            break;
        }
            
        case TRY: {
            b.add(parserLine, TRY); // try bytecode causes a TryMarker to be pushed
            int tryInsn = b.size() - 1;
            // parse the expression to be TRYed
            parseStatement(b, null); 
            // pop the try  marker. this is pushed when the TRY bytecode is executed                              
            b.add(parserLine, POP);
            // jump forward to the end of the catch block, start of the finally block                             
            b.add(parserLine, JMP);                                  
            int successJMPInsn = b.size() - 1;
            
            if (peekToken() != CATCH && peekToken() != FINALLY)
                throw pe("try without catch or finally");
            
            int catchJMPDistance = -1;
            if (peekToken() == CATCH) {
                catchJMPDistance = b.size() - tryInsn;
                String exceptionVar;
                getToken();
                consume(LP);
                consume(NAME);
                exceptionVar = string;
                consume(RP);
                b.add(parserLine, TOPSCOPE);                        // the exception is on top of the stack; put it to the chosen name
                b.add(parserLine, SWAP);
                b.add(parserLine, LITERAL,exceptionVar);
                b.add(parserLine, SWAP);
                b.add(parserLine, PUT);
                b.add(parserLine, POP);
                b.add(parserLine, POP);
                parseStatement(b, null);
                // pop the try and catch markers
                b.add(parserLine,POP);
                b.add(parserLine,POP);
            }
            
            // jump here if no exception was thrown
            b.set(successJMPInsn, new Integer(b.size() - successJMPInsn)); 
                        
            int finallyJMPDistance = -1;
            if (peekToken() == FINALLY) {
                b.add(parserLine, LITERAL, null); // null FinallyData
                finallyJMPDistance = b.size() - tryInsn;
                consume(FINALLY);
                parseStatement(b, null);
                b.add(parserLine,FINALLY_DONE); 
            }
            
            // setup the TRY arguments
            b.set(tryInsn, new int[] { catchJMPDistance, finallyJMPDistance });
            
            break;
        }
            
        case FOR: {
            consume(LP);
            
            tok = getToken();
            boolean hadVar = false;                                      // if it's a for..in, we ignore the VAR
            if (tok == VAR) { hadVar = true; tok = getToken(); }
            String varName = string;
            boolean forIn = peekToken() == IN;                           // determine if this is a for..in loop or not
            pushBackToken(tok, varName);
            
            if (forIn) {
                b.add(parserLine, NEWSCOPE);                             // for-loops always create new scopes
                b.add(parserLine, LITERAL, varName);                     // declare the new variable
                b.add(parserLine, DECLARE);
                    
                b.add(parserLine, LOOP);                                 // we actually only add this to ensure that BREAK works
                b.add(parserLine, POP);                                  // discard the first-iteration indicator
                int size = b.size();
                consume(NAME);
                consume(IN);
                startExpr(b, -1);
                b.add(parserLine, PUSHKEYS);                             // push the keys as an array; check the length
                b.add(parserLine, LITERAL, "length");
                b.add(parserLine, GET);
                consume(RP);
                    
                b.add(parserLine, LITERAL, new Integer(1));              // decrement the length
                b.add(parserLine, SUB);
                b.add(parserLine, DUP);
                b.add(parserLine, LITERAL, new Integer(0));              // see if we've exhausted all the elements
                b.add(parserLine, LT);
                b.add(parserLine, JF, new Integer(2));
                b.add(parserLine, BREAK);                                // if we have, then BREAK
                b.add(parserLine, GET_PRESERVE);                         // get the key out of the keys array
                b.add(parserLine, LITERAL, varName);
                b.add(parserLine, PUT);                                  // write it to this[varName]                          
                parseStatement(b, null);                                 // do some stuff
                b.add(parserLine, CONTINUE);                             // continue if we fall out the bottom

                b.set(size - 1, new Integer(b.size() - size + 1));       // BREAK to here
                b.add(parserLine, OLDSCOPE);                             // restore the scope
                    
            } else {
                if (hadVar) pushBackToken(VAR, null);                    // yeah, this actually matters
                b.add(parserLine, NEWSCOPE);                             // grab a fresh scope
                    
                parseStatement(b, null);                                 // initializer
                CompiledFunctionImpl e2 =                                    // we need to put the incrementor before the test
                    new JS.CompiledFunction(sourceName, parserLine, null, null);  // so we save the test here
                if (peekToken() != SEMI)
                    startExpr(e2, -1);
                else
                    e2.add(parserLine, b.LITERAL, Boolean.TRUE);         // handle the for(foo;;foo) case
                consume(SEMI);
                if (label != null) b.add(parserLine, LABEL, label);
                b.add(parserLine, LOOP);
                int size2 = b.size();
                    
                b.add(parserLine, JT, new Integer(0));                   // if we're on the first iteration, jump over the incrementor
                int size = b.size();
                if (peekToken() != RP) {                                 // do the increment thing
                    startExpr(b, -1);
                    b.add(parserLine, POP);
                }
                b.set(size - 1, new Integer(b.size() - size + 1));
                consume(RP);
                    
                b.paste(e2);                                             // ok, *now* test if we're done yet
                b.add(parserLine, JT, new Integer(2));                   // break out if we don't meet the test
                b.add(parserLine, BREAK);
                parseStatement(b, null);
                b.add(parserLine, CONTINUE);                             // if we fall out the bottom, CONTINUE
                b.set(size2 - 1, new Integer(b.size() - size2 + 1));     // end of the loop
                    
                b.add(parserLine, OLDSCOPE);                             // get our scope back
            }
            break;
        }
                
        case NAME: {  // either a label or an identifier; this is the one place we're not LL(1)
            String possiblyTheLabel = string;
            if (peekToken() == COLON) {      // label
                consume(COLON);
                parseStatement(b, possiblyTheLabel);
                break;
            } else {                         // expression
                pushBackToken(NAME, possiblyTheLabel);  
                startExpr(b, -1);
                b.add(parserLine, POP);
                if ((mostRecentlyReadToken != RC || peekToken() == SEMI) && peekToken() != -1 && mostRecentlyReadToken != SEMI) consume(SEMI);
                break;
            }
        }

        case SEMI: return;                                               // yep, the null statement is valid

        case LC: {  // blocks are statements too
            pushBackToken();
            b.add(parserLine, NEWSCOPE);
            parseBlock(b, label);
            b.add(parserLine, OLDSCOPE);
            break;
        }

        default: {  // hope that it's an expression
            pushBackToken();
            startExpr(b, -1);
            b.add(parserLine, POP);
            if ((mostRecentlyReadToken != RC || peekToken() == SEMI) && peekToken() != -1 && mostRecentlyReadToken != SEMI) consume(SEMI);
            break;
        }
        }
    }


    // ParserException //////////////////////////////////////////////////////////////////////
    private IOException pe(String s) { return new IOException(sourceName + ":" + parserLine + " " + s); }
    
}

